/*******************************************************************************

    C-Sheep Counterpart Library (v1.0)

    Copyright (C) Eike Anderson, Steffen Engel and Leigh McLoughlin
    Copyright (C) NCCA 2006,2007
                  The National Centre for Computer Animation
                  Media School, Bournemouth University
                  Talbot Campus, Fern Barrow, 
                  Poole, Dorset BH12 5BB, UK

    This software is licensed under the Free Academic License ("AFL") v. 3.0
    http://www.opensource.org/licenses/afl-3.0.php

*******************************************************************************/

#include<stdlib.h>
#include<stdio.h>
#include<time.h>

/* if compiled using a microsoft compiler, include windows.h */
#ifdef _MSC_VER
	#include <windows.h>
#endif

/* dependencies: this library must be linked against SDL (SDL and SDLmain), OpenGL & GLU */
#include <SDL/SDL.h>
#include <GL/gl.h>
#include <GL/glu.h>

/* when compiling, make sure that the location of sheep.h is in the include path */
#include<sheep.h>

/*
   note: to build this library to use textures, these must exist in a directory 
   "scld" in the working directory - PRETTY must be defined as a symbolic constant
*/
/* #define PRETTY */

/* global variables */

static char meadow_file[256];
static char meadow_file_set=0;
static int meadow[64][64][3]; /* the level grid */
static char delay=0;
static char cla=0;

/* sheep position */
static int currentX=1;
static int currentY=1;
static int dirX=0;
static int dirY=1;

#ifdef PRETTY
#include<math.h>
#define PI 3.141592653f

/* textures */
static unsigned int grassTexture=0;
static unsigned int hedgeTexture=0;
static unsigned int cobblesTexture=0;
#endif

/* sheep speed */
static int speed=250;

/* width & height */
#ifdef PRETTY
static short int xWidth=800;
static short int yHeight=800;
#else
static short int xWidth=600;
static short int yHeight=600;
#endif

/* utility functions */

static void error(int em)
{
  /* print error message identified by em and exit the program */
  /* errors: */
  switch(em)
  {
    case 1:  fprintf(stderr,"Error (1): no OpenGL rendering context.\n");
             /* 1 - no OpenGL rendering context */
             break;
    case 2:  fprintf(stderr,"Error (2): Meadow could not be loaded.\n");
             /* 2 - level file not found */
             break;
    case 3:  fprintf(stderr,"Error (3): Texture could not be loaded.\n");
             /* 3 - texture not found */
             break;
    default: fprintf(stderr,"Unknown error!\n");
             break;
  }
  exit(EXIT_FAILURE);
}

static void handleKey(SDL_keysym *keysym)
{
  /* keyboard handler - to be called by draw */
  switch(keysym->sym)
  {
    case SDLK_PLUS: /* '+' increases the sheep's speed */
	case SDLK_KP_PLUS:
		if(speed<1000) speed+=50;
		break;
	case SDLK_MINUS: /* '-' decreases the sheep's speed */
	case SDLK_KP_MINUS:
		if(speed>50) speed-=50;
		break;
	case SDLK_ESCAPE:
		exit(EXIT_SUCCESS);
		break;
	case SDLK_F10:
		delay=0;
		exit(EXIT_SUCCESS);
		break;
  }
}

static void draw(int action)
{
  /* rendering function */
  static unsigned int last = 0, current = 0, deltaT = 0;

  int i,j,timer=0;
  double frametime=0.0;

  float sheepX,sheepY,angle,da,dx,dy;
  
  SDL_Event event;
  
  sheepX=(float)(10.0f*currentX)+10.0f;
  sheepY=(float)(10.0f*currentY)+10.0f;
  
  if(dirY==1)
  {
    angle=0.0f;
  }
  else if(dirX==1)
  {
    angle=270.0f;
  }
  else if(dirY==-1)
  {
    angle=180.0f;
  }
  else if(dirX==-1)
  {
    angle=90.0f;
  }

  switch(action)
  {
    case  1: da=0.0f;
             dx=(float)dirX;
             dy=(float)dirY;
             timer=(speed*2);
             break;
    case  2: da=0.0f;
             dx=(float)-dirX;
             dy=(float)-dirY;
             timer=(speed*2);
             break;
    case  3: dx=dy=0.0f;
             da=1.0f;
             timer=speed;
             break;
    case  4: dx=dy=0.0f;
             da=-1.0f;
             timer=speed;
             break;
    default: dx=dy=da=0.0f;
             break;
  }

  do /* render/update loop */
  {
    while( SDL_PollEvent( &event ) )
    {
      switch(event.type)
      {
        case SDL_QUIT:
			delay=0;
			exit(EXIT_SUCCESS);
			break;
		case SDL_KEYUP:
			handleKey(&(event.key.keysym));
			break;
      }
    }

    glClear(GL_COLOR_BUFFER_BIT);
    glLoadIdentity();

#ifdef PRETTY
	/* render background */
	glEnable(GL_TEXTURE_2D);
	glBindTexture(GL_TEXTURE_2D,grassTexture);
	glColor3f(1.0f,1.0f,1.0f);
	glBegin(GL_POLYGON);
	glTexCoord2f(0.0f,0.0f);  glVertex2f(0.0f,0.0f);
	glTexCoord2f(0.0f,8.0f);  glVertex2f(0.0f,660.0f);
	glTexCoord2f(8.0f,8.0f);  glVertex2f(660.0f,660.0f);
	glTexCoord2f(8.0f,0.0f);  glVertex2f(660.0f,0.0f);
	glEnd();
	glDisable(GL_TEXTURE_2D);
#endif

    for(i=0;i<64;i++)
    {
      for(j=0;j<64;j++)
      {
        if(meadow[j][i][1]==255) /* hedge */
        {
#ifdef PRETTY
			glEnable(GL_TEXTURE_2D);
			glBindTexture(GL_TEXTURE_2D,hedgeTexture);
			glColor3f(1.0f,1.0f,1.0f);
#else
          glColor3f(0.0f,0.5f,0.0f);
#endif
          glBegin(GL_POLYGON);
          glTexCoord2f(((float)i)/16.0f,((float)j)/16.0f); glVertex2f((10.0f*(float)i)+10.0f,(10.0f*(float)j)+10.0f);
          glTexCoord2f(((float)i)/16.0f+0.0625f,((float)j)/16.0f); glVertex2f((10.0f*(float)i)+20.0f,(10.0f*(float)j)+10.0f);
          glTexCoord2f(((float)i)/16.0f+0.0625f,((float)j)/16.0f+0.0625f); glVertex2f((10.0f*(float)i)+20.0f,(10.0f*(float)j)+20.0f);
          glTexCoord2f(((float)i)/16.0f,((float)j)/16.0f+0.0625f); glVertex2f((10.0f*(float)i)+10.0f,(10.0f*(float)j)+20.0f);
          glEnd();  
#ifdef PRETTY
		  glDisable(GL_TEXTURE_2D);
#endif
        }
        else if(meadow[j][i][1]>128) /* hedge */
        {
          glColor3f(0.0f,0.9f,0.0f);
          glBegin(GL_POLYGON);
          glVertex2f((10.0f*(float)i)+10.0f,(10.0f*(float)j)+10.0f);
          glVertex2f((10.0f*(float)i)+20.0f,(10.0f*(float)j)+10.0f);
          glVertex2f((10.0f*(float)i)+20.0f,(10.0f*(float)j)+20.0f);
          glVertex2f((10.0f*(float)i)+10.0f,(10.0f*(float)j)+20.0f);
          glEnd();
        }
        else
        {
          if(meadow[j][i][0]==255) /* cobble stones */
          {
#ifdef PRETTY
			glEnable(GL_TEXTURE_2D);
			glBindTexture(GL_TEXTURE_2D,cobblesTexture);
			glColor3f(1.0f,1.0f,1.0f);
#else
            glColor3f(0.4f,0.4f,0.5f);
#endif
            glBegin(GL_POLYGON);
            glTexCoord2f(((float)i)/2.67f,((float)j)/2.67f); glVertex2f((10.0f*(float)i)+10.0f,(10.0f*(float)j)+10.0f);
            glTexCoord2f(((float)i)/2.67f+0.375f,((float)j)/2.67f); glVertex2f((10.0f*(float)i)+20.0f,(10.0f*(float)j)+10.0f);
            glTexCoord2f(((float)i)/2.67f+0.375f,((float)j)/2.67f+0.375f); glVertex2f((10.0f*(float)i)+20.0f,(10.0f*(float)j)+20.0f);
            glTexCoord2f(((float)i)/2.67f,((float)j)/2.67f+0.375f); glVertex2f((10.0f*(float)i)+10.0f,(10.0f*(float)j)+20.0f);
            glEnd();         
#ifdef PRETTY
		  glDisable(GL_TEXTURE_2D);
#endif
          }
          switch(meadow[j][i][2]) /* objects */
          {
            case 255: /* ball */
#ifdef PRETTY
                      {
						  int k=0;
						  float angle=0.0f;
						  glColor3f(1.0f,0.38f,1.0f);
						  glPushMatrix();
						  glTranslatef(10.0f*((float)i)+15.0f,10.0f*((float)j)+15.0f,0.0f);
						  glBegin(GL_POLYGON);
						  for(k=0;k<10;k++)
						  {
							  angle=2.0f*PI*((float)k)/10.0f;
							  glVertex2d(3.0*cos(angle),3.0*sin(angle));
						  }
						  glEnd();
						  glPopMatrix();
					  }
#else
                      glColor3f(1.0f,0.58f,1.0f);
                      glBegin(GL_POLYGON);
                        glVertex2f((10.0f*(float)i)+15.0f,(10.0f*(float)j)+10.0f);
                        glVertex2f((10.0f*(float)i)+20.0f,(10.0f*(float)j)+15.0f);
                        glVertex2f((10.0f*(float)i)+15.0f,(10.0f*(float)j)+20.0f);
                        glVertex2f((10.0f*(float)i)+10.0f,(10.0f*(float)j)+15.0f);
                      glEnd();
#endif
                      break; 
          }
        }
      }
    }

    /* sheep */
    glTranslatef((sheepX+5.0f),(sheepY+5.0f),0.0f);
    glRotatef(angle,0.0f,0.0f,1.0f);
    glColor3f(0.95f,0.95f,0.95f);
#ifdef PRETTY
	{
		int i=0;
		float angle=0.0f;
		glBegin(GL_POLYGON);
		for(i=0;i<10;i++)
		{
			angle=2.0f*PI*((float)i)/10.0f;
			glVertex2d(3.0*cos(angle),3.0*sin(angle)-1.0);
		}
		glEnd();
		glColor3f(0.15f,0.15f,0.15f);
		glBegin(GL_POLYGON);
		for(i=0;i<10;i++)
		{
			angle=2.0f*PI*((float)i)/10.0f;
			glVertex2d(2.0*cos(angle),2.0*sin(angle)+2.0);
		}
		glEnd();
	}
#else
    glBegin(GL_POLYGON);
      glVertex2f(-4.0f,-4.0f);
      glVertex2f(0.0f,4.0f);
      glVertex2f(4.0f,-4.0f);
    glEnd();
#endif

    /* update position */
    frametime=(double)deltaT/1000.0;

    angle+=da*(90.0f*4.0f*(float)frametime);
    sheepX+=(dx*20.0f*(float)frametime);
    sheepY+=(dy*20.0f*(float)frametime);
    SDL_GL_SwapBuffers();    
    current = SDL_GetTicks();
    deltaT = current - last;
	last = current;
    timer-=deltaT;
  }  while(timer>0);
}

static void quit(void)
{
  /* free all allocated data, resources and devices */
  if(delay)
  {
    printf("Please press <RETURN> to quit.\n");
    getchar();
  }
  SDL_Quit();
}

static unsigned int loadTexture(char *filename)
{
	unsigned int texObj=0;
	SDL_Surface *image=NULL;
	
	if(filename==NULL) return 0;
	image=SDL_LoadBMP(filename);
	if(image==NULL) 
	{
		return 0;
	}
	else
	{
		glGenTextures(1,&texObj);
		glBindTexture(GL_TEXTURE_2D,texObj);
		glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_S,GL_REPEAT);
		glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_T,GL_REPEAT);
		glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_LINEAR);
		glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_LINEAR);
		glTexImage2D(GL_TEXTURE_2D,0,image->format->BitsPerPixel/8,image->w,image->h,0,GL_BGR_EXT,GL_UNSIGNED_BYTE,image->pixels);
		SDL_FreeSurface(image);
	}
	return texObj;
}

static int loadMeadow(char *filename)
{
  /* load level & initialise - to be called by checkGL */
  int i,j,k,width,height;
  SDL_Surface *levelSurface=NULL;

  if(filename==NULL) return 0;
  levelSurface=SDL_LoadBMP(filename);
  if(levelSurface==NULL) 
  {
  	return 0;
  }
  else
  {
  	SDL_PixelFormat *fmt;
  	Uint8 *pixel;
  	fmt=levelSurface->format;
  	
  	pixel=(Uint8*)levelSurface->pixels;
        SDL_UnlockSurface(levelSurface);

        width=levelSurface->w;
        height=levelSurface->h;

        if(width!=height) return 0;
        if(width>64) return 0;

        for(i=0;i<64;i++)
        {
	        for(j=0;j<64;j++)
	        {
	            meadow[j][i][0]=0; 
	            meadow[j][i][1]=0;
	            meadow[j][i][2]=0; 
	        }
        }

        for(i=0,k=height-1;i<height;i++,k--)
        {
	        for(j=0;j<width;j++)
	        {
	            Uint8 r=0,g=0,b=0;

	            Uint32 indexR=0,indexG=0,indexB=0;

	            indexB=(3*i)*levelSurface->w+(3*j);
	            indexG=(3*i)*levelSurface->w+(3*j)+1;
	            indexR=(3*i)*levelSurface->w+(3*j)+2;
	            r=pixel[indexR]; /* r is surface */
	            g=pixel[indexG]; /* g is obstacles */
	            b=pixel[indexB]; /* b is objects */

	            meadow[k][j][0]=r; 
	            meadow[k][j][1]=g;
	            meadow[k][j][2]=b; 
	        }
        }

        SDL_FreeSurface(levelSurface);

        for(i=0;i<height;i++)
	{
	        meadow[i][0][1]=255;
	        meadow[i][(width-1)][1]=255;
        }
        for(i=0;i<width;i++)
        {
	        meadow[0][i][1]=255;
	        meadow[(height-1)][i][1]=255;
        }
  }
  return 1;
}

static int checkGL(void)
{
  static int running=0;
  if(!running)
  {
  	int flags = SDL_OPENGLBLIT | SDL_HWSURFACE;
  	short int bpp = 16;
        short int depth_size = 16;
        SDL_Surface *screen = NULL;

  	/* initialise OpenGL */

  	if ( SDL_Init( SDL_INIT_VIDEO | SDL_INIT_TIMER ) < 0 ) return 0;

	atexit(quit);

        SDL_GL_SetAttribute( SDL_GL_DEPTH_SIZE, depth_size );
        SDL_GL_SetAttribute( SDL_GL_DOUBLEBUFFER, 1 );
        SDL_WM_SetCaption("The Meadow",NULL);
        screen = SDL_SetVideoMode(xWidth, yHeight, bpp, flags );
    
        if( screen == NULL ) return 0;

	if(meadow_file_set)
	{
		if(loadMeadow(meadow_file)==0)
		{
			error(2);
		}
	}
	else if(loadMeadow("default.scn")==0)
	{
	  error(2);
	}

#ifdef PRETTY
	if((grassTexture=loadTexture("scld/grass.bmp"))==0)
	{
		error(3);
	}
	if((hedgeTexture=loadTexture("scld/hedge.bmp"))==0)
	{
		error(3);
	}
	if((cobblesTexture=loadTexture("scld/cobbles.bmp"))==0)
	{
		error(3);
	}
#endif
        /* set OpenGL states */
        glViewport(0,0,xWidth,yHeight);
	glMatrixMode(GL_PROJECTION);
        glLoadIdentity();
        gluOrtho2D(0,660,0,660);
        glMatrixMode(GL_MODELVIEW);
        glClearColor(0.0,1.0,0.0,0.0);
  	
  	running=1;
        delay=1;
  }
  return running;
}

/* C-Sheep Counterpart Library API functions */
void initMeadow(int *argc,char **argv)
{
	/* retrieve the level to use from the command-line - similar syntax to glutInit() */
	if((*argc)>1)
	{
		cla=*argc;
		meadow_file_set=1;
		strcpy(meadow_file,argv[1]);
	}
}

/* C-Sheep API functions */
void pause(void)
{
  /* wait for input before program continues */
  int loop=1;
  if(!checkGL()) error(1); /* check if rendering context exists */
  while(loop)
  {
  	/* check for input, set loop to 0 if it comes */
        loop=0;
        draw(0);
  }
}

void initialise(int x,int y)
{
  /* initialise the initial sheep position */
  if(!checkGL()) error(1); /* check if rendering context exists */
  
  currentX=x;
  currentY=y;
  dirX=0;
  dirY=1;
  
  draw(0);
}

void step(void)
{
  /* take a step forwards */
  if(!checkGL()) error(1); /* check if rendering context exists */

  if(meadow[currentY+dirY][currentX+dirX][1]!=255)
  {
    draw(1);
    currentX+=dirX;
    currentY+=dirY;
  }
  else
  {
    fprintf(stdout,"Warning: cannot step onto position %d,%d!\n",currentX+dirX,currentY+dirY);
    draw(0);
  }
}

void backstep(void)
{
  /* take a step back */
  if(!checkGL()) error(1); /* check if rendering context exists */
  
  if(meadow[currentY-dirY][currentX-dirX][1]!=255)
  {
    draw(2);
    currentX-=dirX;
    currentY-=dirY;
  }
  else
  {
    fprintf(stdout,"Warning: cannot backstep onto position %d,%d!\n",currentX+dirX,currentY+dirY);
    draw(0);
  }
}

void turn_left(void)
{
  /* turn left */
  if(!checkGL()) error(1); /* check if rendering context exists */

  draw(3);

  if(dirY==1)
  {
    dirX=-1;
    dirY=0;
  }
  else if(dirX==1)
  {
    dirX=0;
    dirY=1;
  }
  else if(dirY==-1)
  {
    dirX=1;
    dirY=0;
  }
  else if(dirX==-1)
  {
    dirX=0;
    dirY=-1;
  }
}

void turn_right(void)
{
  /* turn right */
  if(!checkGL()) error(1); /* check if rendering context exists */

  draw(4);

  if(dirY==1)
  {
    dirX=1;
    dirY=0;
  }
  else if(dirX==1)
  {
    dirX=0;
    dirY=-1;
  }
  else if(dirY==-1)
  {
    dirX=-1;
    dirY=0;
  }
  else if(dirX==-1)
  {
    dirX=0;
    dirY=1;
  }
}

int blocked(int direction)
{
  /* check if path is blocked in the indicated direction */
  int retval=0;
  if(!checkGL()) error(1); /* check if rendering context exists */

  switch(direction)
  {
    case FRONT: if(dirY==1)
                {
                  if(meadow[currentY+1][currentX][1]>128) retval=1;
                }
                else if(dirX==1)
                {
                  if(meadow[currentY][currentX+1][1]>128) retval=1;
                }
                else if(dirY==-1)
                {
                  if(meadow[currentY-1][currentX][1]>128) retval=1;
                }
                else if(dirX==-1)
                {
                  if(meadow[currentY][currentX-1][1]>128) retval=1;
                }
                break;
    case  BACK: if(dirY==1)
                {
                  if(meadow[currentY-1][currentX][1]>128) retval=1;
                }
                else if(dirX==1)
                {
                  if(meadow[currentY][currentX-1][1]>128) retval=1;
                }
                else if(dirY==-1)
                {
                  if(meadow[currentY+1][currentX][1]>128) retval=1;
                }
                else if(dirX==-1)
                {
                  if(meadow[currentY][currentX+1][1]>128) retval=1;
                }
                break;
    case RIGHT: if(dirY==1)
                {
                  if(meadow[currentY][currentX+1][1]>128) retval=1;
                }
                else if(dirX==1)
                {
                  if(meadow[currentY-1][currentX][1]>128) retval=1;
                }
                else if(dirY==-1)
                {
                  if(meadow[currentY][currentX-1][1]>128) retval=1;
                }
                else if(dirX==-1)
                {
                  if(meadow[currentY+1][currentX][1]>128) retval=1;
                }
                break;
    case  LEFT: if(dirY==1)
                {
                  if(meadow[currentY][currentX-1][1]>128) retval=1;
                }
                else if(dirX==1)
                {
                  if(meadow[currentY+1][currentX][1]>128) retval=1;
                }
                else if(dirY==-1)
                {
                  if(meadow[currentY][currentX+1][1]>128) retval=1;
                }
                else if(dirX==-1)
                {
                  if(meadow[currentY-1][currentX][1]>128) retval=1;
                }
                break;
    default:    fprintf(stdout,"Warning: checking unknown direction (%d)!",direction);
                break;
  }

/*  draw(0); */


  return retval;
}

int found(int object)
{
  /* check if the object indicated has been located at the current position */
  int retval=0;
  if(!checkGL()) error(1); /* check if rendering context exists */

  /* do stuff */
  switch(object)
  {
    case BALL: if(meadow[currentY][currentX][2]==255)
                 retval=1; 
               break;
    default:   break;
  }

/*  draw(0); */


  return retval;
}

float query(int subject)
{
  /* answer the query regarding the indicated subject */
  float retval=1.0;
  if(!checkGL()) error(1); /* check if rendering context exists */

  /* do stuff */
  switch(subject)
  {
	case POSITION_X:
		retval=(float)currentX;
		break;
	case POSITION_Y:
		retval=(float)currentY;
		break;
  }

/*  draw(0); */

  return retval;
}

